/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "entity.h"
#include "assert/assert.h"
#include "util/memory.h"
#include "overworld/overworld.h"

entitycallback_t ENTITY_CALLBACKS[ENTITY_TYPE_COUNT] = {
  { NULL, NULL, NULL },
  { playerInit, playerUpdate, NULL },
  { npcInit, npcUpdate, npcInteract }
};

entity_t ENTITY_TEST;

void entityInit(
  entity_t *ent,
  const entitytype_t type
) {
  assertNotNull(ent, "Entity cannot be NULL");
  assertTrue(type < ENTITY_TYPE_COUNT, "Invalid entity type");

  // Init values
  memoryZero(ent, sizeof(entity_t));
  ent->type = type;

  // Call init
  assertNotNull(ENTITY_CALLBACKS[type].init, "Entity type init callback err.");
  ENTITY_CALLBACKS[type].init(ent);

  entityTurn(ent, FACING_DIRECTION_SOUTH);
}

void entityUpdate(entity_t *ent) {
  assertNotNull(ent, "Entity cannot be NULL");
  assertTrue(ent->type < ENTITY_TYPE_COUNT, "Invalid entity type");
  assertNotNull(ENTITY_CALLBACKS[ent->type].update, "Entity type update err.");
  ENTITY_CALLBACKS[ent->type].update(ent);

  // Handle moving
  if(ent->subX > 0) {
    ent->subX -= ENTITY_MOVE_SPEED;
    if(ent->subX < 0) ent->subX = 0;
  } else if(ent->subX < 0) {
    ent->subX += ENTITY_MOVE_SPEED;
    if(ent->subX > 0) ent->subX = 0;
  } else if(ent->subY > 0) {
    ent->subY -= ENTITY_MOVE_SPEED;
    if(ent->subY < 0) ent->subY = 0;
  } else if(ent->subY < 0) {
    ent->subY += ENTITY_MOVE_SPEED;
    if(ent->subY > 0) ent->subY = 0;
  }
}

void entityTurn(entity_t *ent, const facingdir_t dir) {
  assertNotNull(ent, "Entity cannot be NULL");
  assertTrue(dir < FACING_DIRECTION_COUNT, "Invalid facing direction");

  ent->direction = dir;
  switch(dir) {
    case FACING_DIRECTION_SOUTH:
      ent->frame = 25;
      break;
    case FACING_DIRECTION_NORTH:
      ent->frame = 24;
      break;
    case FACING_DIRECTION_EAST:
      ent->frame = 26;
      break;
    case FACING_DIRECTION_WEST:
      ent->frame = 27;
      break;
  }
}

void entityMove(entity_t *ent, const facingdir_t dir) {
  assertNotNull(ent, "Entity cannot be NULL");
  if(entityIsMoving(ent)) return;

  entityTurn(ent, dir);

  uint8_t targetX = ent->x, targetY = ent->y;
  facingDirAdd(ent->direction, &targetX, &targetY);

  // Check oob
  if(targetX < 0 || targetX >= OVERWORLD.map.width) return;
  if(targetY < 0 || targetY >= OVERWORLD.map.height) return;

  // Check tile at target
  uint8_t i = 0;
  tileid_t tileId;
  for(i = 0; i < OVERWORLD.map.layerCount; i++) {
    tileId = mapGetTileIdAtPosition(&OVERWORLD.map, i, targetX, targetY);
    if(tileIsSolid(tileId)) return;
  }

  // Check for entity at target
  entity_t *atPos = overworldEntityGetAtPosition(targetX, targetY);
  if(atPos != NULL) return;

  // Commit to move
  ent->x = targetX;
  ent->y = targetY;
  
  switch(dir) {
    case FACING_DIRECTION_EAST:
      ent->subX = -OVERWORLD_ENTITY_WIDTH;
      break;
    case FACING_DIRECTION_WEST:
      ent->subX = OVERWORLD_ENTITY_WIDTH;
      break;
    case FACING_DIRECTION_NORTH:
      ent->subY = -OVERWORLD_ENTITY_HEIGHT;
      break;
    case FACING_DIRECTION_SOUTH:
      ent->subY = OVERWORLD_ENTITY_HEIGHT;
      break;
  }
}

bool_t entityIsMoving(const entity_t *ent) {
  assertNotNull(ent, "Entity cannot be NULL");
  return ent->subX != 0 || ent->subY != 0;
}

bool_t entityInteract(entity_t *ent) {
  assertNotNull(ent, "Entity cannot be NULL.");
  assertTrue(ent->type == ENTITY_TYPE_PLAYER, "Entity should be player.");
  if(entityIsMoving(ent)) return false;

  uint8_t targetX = ent->x, targetY = ent->y;
  facingDirAdd(ent->direction, &targetX, &targetY);

  entity_t *target = overworldEntityGetAtPosition(targetX, targetY);
  assertFalse(target == ent, "Cannot interact with self.");

  if(target == NULL) return false;
  assertTrue(target->type < ENTITY_TYPE_COUNT, "Invalid entity type.");

  if(ENTITY_CALLBACKS[target->type].interact == NULL) return false;
  ENTITY_CALLBACKS[target->type].interact(ent, target);
  
  return true;
}