/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "facing.h"
#include "player.h"
#include "npc.h"

typedef enum {
  ENTITY_TYPE_NULL,
  ENTITY_TYPE_PLAYER,
  ENTITY_TYPE_NPC
} entitytype_t;

#define ENTITY_TYPE_COUNT 3

typedef struct {
  void (*init)(entity_t *);
  void (*update)(entity_t *);
  void (*interact)(entity_t *, entity_t *);
} entitycallback_t;

typedef struct _entity_t {
  uint8_t x, y;
  int8_t subX, subY;
  uint8_t frame;
  facingdir_t direction;
  entitytype_t type;

  union {
    player_t player;
  };
} entity_t;

#define ENTITY_MOVE_SPEED 3

/**
 * Initializes an entity.
 * 
 * @param entity The entity to initialize.
 * @param type The type of entity to initialize.
 */
void entityInit(
  entity_t *entity,
  const entitytype_t type
);

/**
 * Updates an entity.
 * 
 * @param entity The entity to update.
 */
void entityUpdate(entity_t *entity);

/**
 * Turns an entity in a given direction.
 * 
 * @param entity The entity to turn.
 * @param dir The direction to turn the entity.
 */
void entityTurn(entity_t *entity, const facingdir_t dir);

/**
 * Moves an entity in a given direction.
 * 
 * @param entity The entity to move.
 * @param dir The direction to move the entity.
 */
void entityMove(entity_t *entity, const facingdir_t dir);

/**
 * Checks if an entity is moving.
 * 
 * @param entity The entity to check.
 * @return TRUE if the entity is moving, FALSE otherwise.
 */
bool_t entityIsMoving(const entity_t *entity);

/**
 * Make this entity attempt to interact with the world.
 * 
 * @param entity The entity that is doing the interaction.
 * @return TRUE if an entity was interacted with, FALSE otherwise.
 */
bool_t entityInteract(entity_t *entity);