/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "framebuffer.h"
#include "assert/assert.h"
#include "assert/assertgl.h"
#include "display/window.h"

framebuffer_t FRAMEBUFFER;

void frameBufferInit(const int32_t width, const int32_t height) {
  assertTrue(width > 0, "Width must be greater than 0");
  assertTrue(height > 0, "Height must be greater than 0");

  FRAMEBUFFER.id = -1;
  FRAMEBUFFER.width = -1;
  FRAMEBUFFER.height = -1;
  FRAMEBUFFER.texture = -1;

  frameBufferSetSize(width, height);
}

void frameBufferSetSize(const int32_t width, const int32_t height) {
  assertTrue(width > 0, "Width must be greater than 0");
  assertTrue(height > 0, "Height must be greater than 0");

  if(FRAMEBUFFER.width == width && FRAMEBUFFER.height == height) return;

  FRAMEBUFFER.width = width;
  FRAMEBUFFER.height = height;

  frameBufferUnbind();

  // Delete old texture and depth buffer
  if(FRAMEBUFFER.texture != -1) {
    glDeleteTextures(1, &FRAMEBUFFER.texture);
    assertNoGLError();
    FRAMEBUFFER.texture = -1;
  }

  if(FRAMEBUFFER.id != -1) {
    glDeleteFramebuffers(1, &FRAMEBUFFER.id);
    assertNoGLError();
    FRAMEBUFFER.id = -1;
  }

  // Generate framebuffer
  glGenFramebuffers(1, &FRAMEBUFFER.id);
  assertNoGLError();
  glBindFramebuffer(GL_FRAMEBUFFER, FRAMEBUFFER.id);
  assertNoGLError();

  // Create new texture
  glGenTextures(1, &FRAMEBUFFER.texture);
  assertNoGLError();
  glBindTexture(GL_TEXTURE_2D, FRAMEBUFFER.texture);
  assertNoGLError();
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
  assertNoGLError();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  assertNoGLError();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  assertNoGLError();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  assertNoGLError();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  assertNoGLError();
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, FRAMEBUFFER.texture, 0);
  assertNoGLError();

  // Check framebuffer completeness
  GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
  assertNoGLError();
  assertTrue(status == GL_FRAMEBUFFER_COMPLETE, "Framebuffer is not complete");

  glBindFramebuffer(GL_FRAMEBUFFER, 0);
  assertNoGLError();
}

void frameBufferBind() {
  if(FRAMEBUFFER.id == -1) return;

  glBindFramebuffer(GL_FRAMEBUFFER, FRAMEBUFFER.id);
  assertNoGLError();

  glViewport(0, 0, FRAMEBUFFER.width, FRAMEBUFFER.height);
  assertNoGLError();

  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
  assertNoGLError();

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  assertNoGLError();
}

void frameBufferTextureBind(const GLuint slot) {
  if(FRAMEBUFFER.texture == -1) return;

  glActiveTexture(GL_TEXTURE0 + slot);
  assertNoGLError();

  glBindTexture(GL_TEXTURE_2D, FRAMEBUFFER.texture);
  assertNoGLError();
}

void frameBufferUnbind() {
  glBindFramebuffer(GL_FRAMEBUFFER, 0);
  assertNoGLError();

  glViewport(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
  assertNoGLError();

  glClearColor(0.05f, 0.05f, 0.05f, 1.0f);
  assertNoGLError();

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  assertNoGLError();
}

void frameBufferDispose() {
  glDeleteFramebuffers(1, &FRAMEBUFFER.id);
  glDeleteTextures(1, &FRAMEBUFFER.texture);

  FRAMEBUFFER.id = -1;
  FRAMEBUFFER.texture = -1;
}