/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "shader.h"
#include "assert/assert.h"
#include "assert/assertgl.h"

void shaderInit(
  shader_t *shader,
  const char *vertexSource,
  const char *fragmentSource
) {
  assertNotNull(shader, "shader must not be NULL");
  assertNotNull(vertexSource, "vertexSource must not be NULL");
  assertNotNull(fragmentSource, "fragmentSource must not be NULL");

  int32_t success;
  char infoLog[SHADER_LOG_LENGTH];

  // Create vertex shader
  shader->vertexShader = glCreateShader(GL_VERTEX_SHADER);
  assertNoGLError();

  glShaderSource(shader->vertexShader, 1, &vertexSource, NULL);
  assertNoGLError();

  glCompileShader(shader->vertexShader);
  assertNoGLError();

  glGetShaderiv(shader->vertexShader, GL_COMPILE_STATUS, &success);
  assertNoGLError();

  if(!success) {
    glGetShaderInfoLog(
      shader->vertexShader, SHADER_LOG_LENGTH, NULL, infoLog
    );
    assertNoGLError();
    assertUnreachable(infoLog);
    return;
  }

  // Create fragment shader
  shader->fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
  assertNoGLError();

  glShaderSource(shader->fragmentShader, 1, &fragmentSource, NULL);
  assertNoGLError();

  glCompileShader(shader->fragmentShader);
  assertNoGLError();

  glGetShaderiv(shader->fragmentShader, GL_COMPILE_STATUS, &success);
  assertNoGLError();

  if(!success) {
    glGetShaderInfoLog(
      shader->fragmentShader, SHADER_LOG_LENGTH, NULL, infoLog
    );
    assertNoGLError();
    assertUnreachable(infoLog);
    return;
  }

  // Create shader program
  shader->shaderProgram = glCreateProgram();
  assertNoGLError();

  glAttachShader(shader->shaderProgram, shader->vertexShader);
  assertNoGLError();

  glAttachShader(shader->shaderProgram, shader->fragmentShader);
  assertNoGLError();

  glLinkProgram(shader->shaderProgram);
  assertNoGLError();

  glGetProgramiv(shader->shaderProgram, GL_LINK_STATUS, &success);
  assertNoGLError();

  if(!success) {
    glGetProgramInfoLog(shader->shaderProgram, SHADER_LOG_LENGTH, NULL, infoLog);
    assertNoGLError();
    assertUnreachable(infoLog);
    return;
  }
}

void shaderUse(const shader_t *shader) {
  assertNotNull(shader, "shader must not be NULL");

  glUseProgram(shader->shaderProgram);
  assertNoGLError();
}

GLuint shaderGetUniform(const shader_t *shader, const char_t *name) {
  assertNotNull(shader, "shader must not be NULL");
  assertNotNull(name, "name must not be NULL");

  GLuint uniform = glGetUniformLocation(shader->shaderProgram, name);
  assertNoGLError();

  return uniform;
}

GLuint shaderGetBlock(const shader_t *shader, const char *name) {
  assertNotNull(shader, "shader must not be NULL");
  assertNotNull(name, "name must not be NULL");

  GLuint blockIndex = glGetUniformBlockIndex(shader->shaderProgram, name);
  assertNoGLError();

  // TODO: I really don't think this should be here at all.
  glUniformBlockBinding(shader->shaderProgram, blockIndex, blockIndex);
  assertNoGLError();

  return blockIndex;
}

void shaderSetTextures(
  const GLuint uniform,
  const GLuint *textures,
  const uint8_t count
) {
  assertNotNull(textures, "textures must not be NULL");
  assertTrue(count > 0, "count must be greater than 0");

  glUniform1iv(uniform, count, textures);
  assertNoGLError();
}

void shaderDispose(shader_t *shader) {
  assertNotNull(shader, "shader must not be NULL");

  glDeleteProgram(shader->shaderProgram);
  assertNoGLError();

  glDeleteShader(shader->vertexShader);
  assertNoGLError();

  glDeleteShader(shader->fragmentShader);
  assertNoGLError();
}