/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "duskgl.h"

#define SHADER_LOG_LENGTH 512

typedef struct {
  GLuint shaderProgram;
  GLuint vertexShader;
  GLuint fragmentShader;
} shader_t;

/**
 * Initializes a shader.
 * 
 * @param shader The shader to initialize.
 * @param vertexSource The vertex shader source.
 * @param fragmentSource The fragment shader source.
 */
void shaderInit(
  shader_t *shader,
  const char_t *vertexSource,
  const char_t *fragmentSource
);

/**
 * Uses a shader.
 * 
 * @param shader The shader to use.
 */
void shaderUse(const shader_t *shader);

/**
 * Gets a uniform from a shader.
 * 
 * @param shader The shader to get the uniform from.
 * @param name The name of the uniform.
 * @return The uniform.
 */
GLuint shaderGetUniform(const shader_t *shader, const char_t *name);

/**
 * Gets a block id from a shader.
 * 
 * @param shader The shader to get the block from.
 * @param name The name of the block.
 * @return The block index/identifier.
 */
GLuint shaderGetBlock(const shader_t *shader, const char_t *name);

/**
 * Sets texture uniforms to a shader.
 * 
 * @param uniform The uniform to set.
 * @param textures The textures to set.
 * @param count The number of textures to set.
 */
void shaderSetTextures(
  const GLuint uniform,
  const GLuint *textures,
  const uint8_t count
);

/**
 * Disposes of a shader.
 * 
 * @param shader The shader to dispose of.
 */
void shaderDispose(shader_t *shader);